//Core bitsensor interface
var ApiConnector = require('./api_connector');
var Config = require('./config');
var Datapoint = require('./datapoint');
var RWLock = require('rwlock');
var lock = new RWLock();

/**
 * Entry point for starting the BitSensor Web Application Security plugin.
 *
 * @constructor
 */
var BitSensor = function () {

  /**
   * @type {Handler[]}
   */
  this.preHandlers = [];
  this.postHandlers = [];
  this.session = require("continuation-local-storage").createNamespace("bitsensor");
  this.isDisabled;
  this.isHandlingRequest;
};

BitSensor.prototype.runSession = function (callback) {
  this.session.run(() => {
    if (!this.isInitialized()) {
      this.initialize();
    }
    this.traceIncrement();
    if (typeof callback === "function")
      callback();
    this.ship();
  });
};

BitSensor.prototype.initialize = function () {
  if (this.isInitialized()) {
    return;
  }

  this.setDatapoint(new Datapoint());
  this.initializeEnabled();
  this.runPreHandlers();
};

BitSensor.prototype.isInitialized = function () {
  return this.session && !!this.session.get('datapoint');
};

BitSensor.prototype.apply = function (consumer) {
  if (this.hasBeenDisabled()) {
    return;
  }

  if (!this.isInitialized()) {
    this.runSession(() => {
      consumer.call(this, this.getDatapoint());
    });
    return;
  }

  consumer.call(this, this.getDatapoint());
};


BitSensor.prototype.getDatapoint = function () {
  return bitSensor.session.get('datapoint');
};

BitSensor.prototype.setDatapoint = function (datapoint) {
  this.session.set("datapoint", datapoint);
};

BitSensor.prototype.ship = function () {
  if (this.hasBeenDisabled())
    return;

  if (this.traceDecrement() > 0) {
    return;
  }

  this.runPostHandlers();
  this.log();
  this.clear();
};

BitSensor.prototype.log = function () {
  if (!this._config) {
    return;
  }
  new ApiConnector().from(this._config.user, this._config.apiKey)
    .to(this._config.uri)
    .with(this.getDatapoint())
    .post(ApiConnector.ACTION_LOG)
    .send();
};

BitSensor.prototype.clear = function () {
  this.isDisabled = null;
  this.isHandlingRequest = null;
  this.session.set('datapoint', null);
};

BitSensor.prototype.hasBeenDisabled = function () {
  if (this.isDisabled == null) {
    return false;
  }

  return this.isDisabled > 0;
};

BitSensor.prototype.setDisabled = function () {
  this.initializeEnabled();
  lock.writeLock((release) => {
    if (this.isDisabled == null) {
      this.isDisabled = 0;
    }
    this.isDisabled++;
    release();
  });

  return this.isDisabled;
};

BitSensor.prototype.setEnabled = function () {
  this.initializeEnabled();
  lock.writeLock((release) => {
    if (this.isDisabled > 0) {
      this.isDisabled--;
    }
    release();
  });

  return this.isDisabled;
};

BitSensor.prototype.initializeEnabled = function () {
  lock.writeLock((release) => {
    if (this.isDisabled == null) {
      this.isDisabled = 0;
    }
    if (this.isHandlingRequest == null) {
      this.isHandlingRequest = 0;
    }
    release();
  });
};

BitSensor.prototype.traceIncrement = function () {
  if (this.isHandlingRequest == null) {
    this.initialize();
  }
  lock.writeLock(function (release) {
    bitSensor.isHandlingRequest++;
    release();
  });
};

BitSensor.prototype.traceDecrement = function () {
  if (this.isHandlingRequest == null) {
    this.initialize();
  }
  lock.writeLock(function (release) {
    if (bitSensor.isHandlingRequest > 0) {
      bitSensor.isHandlingRequest--;
    }
    release();
  });
  return this.isHandlingRequest;
};


BitSensor.prototype.addPreHandler = function (handler) {
  this.preHandlers.push(handler);
};

BitSensor.prototype.addPostHandler = function (handler) {
  this.postHandlers.push(handler);
};


BitSensor.prototype.runPreHandlers = function () {
  this.preHandlers.forEach((handler) => {
    handler.handle(this.getDatapoint());
  });
};

BitSensor.prototype.runPostHandlers = function () {
  this.postHandlers.forEach((handler) => {
    handler.handle(this.getDatapoint());
  });
};

BitSensor.prototype.config = function (config) {
  this._config = new Config(config);
}

module.exports = BitSensor;